/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2008 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Ilia Alshanetsky <iliaa@php.net>                            |
  |          Graham Kelly <graham@php.net>                               |
  +----------------------------------------------------------------------+
*/
/* $Id: optimize.h,v 1.18 2009/04/30 03:32:27 myang Exp $ */

#ifndef OPTIMIZE_H
#define OPTIMIZE_H

#if PHP_WIN32
#  ifndef _W64
#    if !defined(__midl) && (defined(_X86_) || defined(_M_IX86)) && _MSC_VER >= 1300
#       define _W64 __w64
#    else
#       define _W64
#    endif
#  endif
#  ifdef _WIN64
     typedef __int64           intptr_t;
     typedef unsigned __int64  uintptr_t;
#  else
     typedef _W64 int               intptr_t;
     typedef _W64 unsigned int      uintptr_t;
#  endif
#else
#  include <stdint.h>
#endif


/* Op row properties */
#define EXT_UNUSED     0x0ff00
#define EXT_STD        0x00000
#define EXT_OPLINE     0x00100
#define EXT_FCALL      0x00200
#define EXT_ARG        0x00300
#define EXT_SEND       0x00400
#define EXT_CAST       0x00500
#define EXT_INIT_FCALL 0x00600
#define EXT_FETCH      0x00700
#define EXT_DECLARE    0x00800
#define EXT_SEND_NOREF 0x00900
#define EXT_FCLASS     0x00a00
#define EXT_IFACE      0x00b00
#define EXT_ISSET      0x00c00
#define EXT_BIT        0x00d00
#define EXT_CLASS      0x00e00
#define EXT_ASSIGN     0x00f00
#define EXT_FE         0x01000

#define OP1_MASK       0x000f0
#define OP1_UNUSED     0x000f0
#define OP1_STD        0x00000
#define OP1_OPLINE     0x00010
#define OP1_ARG        0x00020
#define OP1_BRK        0x00030
#define OP1_CONT       0x00040
#define OP1_JMPADDR    0x00050
#define OP1_CLASS      0x00060
#define OP1_VAR        0x00070
#define OP1_TMP        0x00080
#define OP1_UCLASS     0x00090

#define OP2_MASK       0x0000f
#define OP2_UNUSED     0x0000f
#define OP2_STD        0x00000
#define OP2_OPLINE     0x00001
#define OP2_FETCH      0x00002
#define OP2_INCLUDE    0x00003
#define OP2_ARG        0x00004
#define OP2_ISSET      0x00005
#define OP2_JMPADDR    0x00006
#define OP2_CLASS      0x00007
#define OP2_VAR        0x00008
#define OP2_TMP        0x00009

#define RES_MASK       0xf0000
#define RES_UNUSED     0xf0000
#define RES_STD        0x00000
#define RES_CLASS      0x10000
#define RES_TMP        0x20000
#define RES_VAR        0x30000

#define OPS_STD       EXT_STD | OP1_STD | OP2_STD | RES_STD


#define OP_OP1 1
#define OP_OP2 2
#define OP_RES 4

/*
 * Special optimizer versions of the zval types 
 *   This is needed because there are many places in the optimizer where we 
 *   need to associate more then one data type with a variable.
 */
#define OPTIMIZER_IS_NULL           1
#define OPTIMIZER_IS_LONG           2
#define OPTIMIZER_IS_DOUBLE         4
#define OPTIMIZER_IS_BOOL           8
#define OPTIMIZER_IS_ARRAY          16
#define OPTIMIZER_IS_OBJECT         32
#define OPTIMIZER_IS_STRING         64
#define OPTIMIZER_IS_RESOURCE       128
#define OPTIMIZER_IS_CONSTANT       256
#define OPTIMIZER_IS_CONSTANT_ARRAY 512
#define OPTIMIZER_IS_UNICODE        1024

#define OPTIMIZER_IS_BOOL_F         2048 /* used for return types... to hint that it might be a specific bool value */
#define OPTIMIZER_IS_BOOL_T         4096

typedef unsigned short optimizer_dt; /* Data type for holding multiple zval data types */


/* op code definitions */
extern unsigned int zops[];

#define MAX_ZOPS sizeof (zops) / sizeof (unsigned int)

#define VAR_NUM(var) ((unsigned int)(((temp_variable *)((intptr_t) var)) - ((temp_variable *)NULL)))
#define VAR_VAL(var) ((unsigned int)((var) * sizeof(temp_variable)))

typedef struct _CBchain {
	struct _CBchain	*next;
	struct _CB	*cb;
} CBchain;

typedef struct _CB {
	zend_op		*start;
	int		len;

	zend_bool	used;
	zend_bool	exn;

	struct _CB	*jmp_1;
	struct _CB	*jmp_2;
	struct _CB	*jmp_ext;
	struct _CB	*jmp_tc;

	struct _CB	*follow;
	struct _CB	*next;

	struct _CBchain	*pchain;
} CB;


#define OFE_TYPE_CONSTRUCTOR        1
#define OFE_TYPE_DESTRUCTOR         2
#define OFE_TYPE_CLONE              3
#define OFE_TYPE___GET              4
#define OFE_TYPE___SET              5
#define OFE_TYPE___UNSET            6
#define OFE_TYPE___ISSET            7
#define OFE_TYPE___CALL             8
#define OFE_TYPE___CALLSTATIC       9
#define OFE_TYPE___TOSTRING        10
#define OFE_TYPE_SERIALIZE_FUNC    11
#define OFE_TYPE_UNSERIALIZE_FUNC  12
#define OFE_TYPE___AUTOLOAD        13

typedef struct _optimizer_class optimizer_class;

typedef struct _optimizer_function {
	zend_function* fe;
	zend_op_array* op_array;

	optimizer_class* class_entry;

	zend_uchar fe_type;

	CB* cfg;
	uint* jmp_lines;

	zend_uchar pass;
} optimizer_function;


struct _optimizer_class {
	zend_class_entry* ce;

	zend_uint fc;                 /* Functions */
	zend_uint fc_static;
	zend_uint fc_private;
	zend_uint fc_private_static;

	zend_uint pc;                 /* Properties */
	zend_uint pc_static;
	zend_uint pc_private;
	zend_uint pc_private_static;

	zend_uint constants_count;

	optimizer_function* constructor;
	optimizer_function* destructor;
	optimizer_function* clone;
	optimizer_function* __get;
	optimizer_function* __set;
	optimizer_function* __unset;
	optimizer_function* __isset;
	optimizer_function* __call;
#if PHP_VERSION_ID >= 50300
	optimizer_function* __callstatic;
#endif
	optimizer_function* __tostring;
	optimizer_function* serialize_func;
	optimizer_function* unserialize_func;

	optimizer_function** function_table;
};

typedef struct _optimizer_source_file2 {
	HashTable main_constants;

	optimizer_function* function_table;
	zend_uint fc;
	zend_uint fc_class;
	zend_uint fc_deleteable;

	zend_uint total_op_count;

	optimizer_class* class_table;
	zend_uint class_count;
	optimizer_function** _class_function_table;

	HashTable fh;
} optimizer_source_file;


#define OP1_TYPE(op) (op)->op1.op_type
#define OP2_TYPE(op) (op)->op2.op_type
#define RES_TYPE(op) (op)->result.op_type

#define __OP1_VAL(op) (op)->op1.u.constant
#define __OP2_VAL(op) (op)->op2.u.constant
#define __RES_VAL(op) (op)->result.u.constant

#define OP1_VALT(op) __OP1_VAL((op)).type
#define OP2_VALT(op) __OP2_VAL((op)).type
#define RES_VALT(op) __OP2_VAL((op)).type

#define OP1_VAL_S(op) Z_STRVAL(__OP1_VAL((op)))
#define OP2_VAL_S(op) Z_STRVAL(__OP2_VAL((op)))

#define OP1_VAL_SL(op) Z_STRLEN(__OP1_VAL((op)))
#define OP2_VAL_SL(op) Z_STRLEN(__OP2_VAL((op)))

#define OP1_VAL_L(op) Z_LVAL(__OP1_VAL((op)))
#define OP2_VAL_L(op) Z_LVAL(__OP2_VAL((op)))

#define OP1_VAL_D(op) Z_DVAL(__OP1_VAL((op)))
#define OP2_VAL_D(op) Z_DVAL(__OP2_VAL((op)))

#define OP1_VAL_B(op) Z_BVAL(__OP1_VAL((op)))
#define OP2_VAL_B(op) Z_BVAL(__OP2_VAL((op)))

#define SET_TO_NOP(op) (op)->opcode = ZEND_NOP; OP1_TYPE((op)) = OP2_TYPE((op)) = RES_TYPE((op)) = IS_UNUSED;

#define SET_TO_NOP_EX(op)            \
	if (OP1_TYPE(op) == IS_CONST) {  \
		zval_dtor(&__OP1_VAL(op));   \
	}                                \
	if (OP2_TYPE(op) == IS_CONST) {  \
		zval_dtor(&__OP2_VAL(op));   \
	}                                \
	SET_TO_NOP(op);                  \

#define RESULT_TYPE(op) (op)->result.u.EA.type
#define USED_RESULT(op) (RESULT_TYPE((op)) & EXT_TYPE_UNUSED)

#define FETCH_TYPE(op) ((op)->op2.u.EA.type)

#define OP1_VR(op) (op)->op1.u.var
#define OP2_VR(op) (op)->op2.u.var
#define RES_VR(op) (op)->result.u.var 

#define OPARRAY_LEN(o) (o)->last

zend_op* optimize_next_op(zend_op* op);
zend_op* optimize_prev_op(zend_op* op);
#define NEXT_OP(op) optimize_next_op(op)
#define PREV_OP(op) optimize_prev_op(op)

/* Operand is variable type (IS_VAR, IS_CV, or IS_TMP_VAR) */
#define OP1_IS_VART(op) (OP1_TYPE(op) == IS_VAR || OP1_TYPE(op) == IS_CV || OP1_TYPE(op) == IS_TMP_VAR)
#define OP2_IS_VART(op) (OP2_TYPE(op) == IS_VAR || OP2_TYPE(op) == IS_CV || OP2_TYPE(op) == IS_TMP_VAR)


/* optimize_to_*_ex flags */
#define OPTIMIZE_TO_DEL_OP    1  /* Delete the 'op' op   */
#define OPTIMIZE_TO_DEL_PREV  2  /* Delete the 'prev' op */
#define OPTIMIZE_TO_FCALL     4  /* Use can_optimize_fcall as opposed to get_next_znode_pos */

#define OPTIMIZE_TO_FULL (OPTIMIZE_TO_DEL_OP | OPTIMIZE_TO_DEL_PREV | OPTIMIZE_TO_FCALL)

void optimize_to_bool_ex(zend_constant* c, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_true_ex(zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_false_ex(zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_null_ex(zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_string_ex(char* str, int str_len, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_long_ex(long val, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_double_ex(double val, zend_op** a, zend_op** b, CB* cbl, zend_op** Ts, zend_uchar flags TSRMLS_DC);
void optimize_to_var_ex(zend_uint var, int var_type,  zend_op **a, zend_op **b, CB *cbl, zend_op **Ts, zend_uchar flags TSRMLS_DC);

/* optimize_to_* macros */
#define optimize_to_bool(c, a, b, cbl, Ts)              optimize_to_bool_ex(c, a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_true(a, b, cbl, Ts)                 optimize_to_true_ex(a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_false(a, b, cbl, Ts)                optimize_to_false_ex(a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_null(a, b, cbl, Ts)                 optimize_to_null_ex(a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_string(str, str_len, a, b, cbl, Ts) optimize_to_string_ex(str, str_len, a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_long(val, a, b, cbl, Ts)            optimize_to_long_ex(val, a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_double(val, a, b, cbl, Ts)          optimize_to_double_ex(val, a, b, cbl, Ts, 0 TSRMLS_CC)
#define optimize_to_var(var, vart, a, b, cbl, Ts)       optimize_to_var_ex(var, vart, a, b, cbl, Ts, 0 TSRMLS_CC)

#define optimize_to_is_cv(var, a, b, cbl, Ts)           optimize_to_var(var, IS_CV, a, b, cbl, Ts)
#define optimize_to_is_tmp_var(var, a, b, cbl, Ts)      optimize_to_var(var, IS_TMP_VAR, a, b, cbl, Ts)
#define optimize_to_is_var(var, a, b, cbl, Ts)          optimize_to_var(var, IS_VAR, a, b, cbl, Ts)

#define OPTIMIZE_FCALL_TO_FLAGS(argc) (OPTIMIZE_TO_FCALL | OPTIMIZE_TO_DEL_OP | (argc > 0 ? OPTIMIZE_TO_DEL_PREV : 0))

#define optimize_fcall_to_bool(c, a, b, argc, cbl, Ts)              optimize_to_bool_ex(c, a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)
#define optimize_fcall_to_true(a, b, argc, cbl, Ts)                 optimize_to_true_ex(a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)
#define optimize_fcall_to_false(a, b, argc, cbl, Ts)                optimize_to_false_ex(a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)
#define optimize_fcall_to_null(a, b, argc, cbl, Ts)                 optimize_to_null_ex(a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)
#define optimize_fcall_to_string(str, str_len, a, b, argc, cbl, Ts) optimize_to_string_ex(str, str_len, a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)
#define optimize_fcall_to_long(val, a, b, argc, cbl, Ts)            optimize_to_long_ex(val, a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)
#define optimize_fcall_to_double(val, a, b, argc, cbl, Ts)          optimize_to_double_ex(val, a, b, cbl, Ts, OPTIMIZE_FCALL_TO_FLAGS(argc) TSRMLS_CC)

#define OPTIMIZE_TO_BOOL(val)	                         \
	if (val) {	                                         \
		optimize_to_true(&prev, &op, cb, Ts);       	 \
	} else {	                                         \
		optimize_to_false(&prev, &op, cb, Ts);           \
	}	                                                 \


#define OPTIMIZE_FCALL_TO_BOOL(val, argc)                  \
	if (val) {                                             \
		optimize_fcall_to_true(&prev, &op, argc, cb, Ts);  \
	} else {                                               \
		optimize_fcall_to_false(&prev, &op, argc, cb, Ts); \
	}                                                      \

#define OPTIMIZE_SAVE(op_array, op) \
do { \
	saved_in_compilation = CG(in_compilation); \
	saved_compiled_filename = CG(compiled_filename); \
	saved_zend_lineno = CG(zend_lineno); \
	CG(in_compilation) = 1; \
	CG(compiled_filename) = (op_array)->filename; \
	CG(zend_lineno) = (op) ? (int)(op)->lineno : 0; \
} while (0)

#define OPTIMIZE_RESTORE(op_array, op) \
do { \
	CG(in_compilation) = saved_in_compilation; \
	CG(compiled_filename) = saved_compiled_filename; \
	CG(zend_lineno) = saved_zend_lineno; \
} while (0)

#define OPTIMIZE_ERROR(type, format, ...) \
do { \
	zend_bool saved_in_compilation; \
	char *saved_compiled_filename; \
	int saved_zend_lineno; \
	OPTIMIZE_SAVE(op_array, op); \
	zend_error(type, format, ##__VA_ARGS__); \
	OPTIMIZE_RESTORE(op_array, op); \
} while (0)

#define OPTIMIZE_ASSERT(x) \
do { \
	if (!(x)) { \
		zend_op *op = NULL; \
		zend_op_array *op_array = OPTIMIZER_G(op_array); \
		OPTIMIZE_ERROR(E_ERROR, "Assertion '" #x "' failed"); \
	} \
} while (0)

int optimize_get_constant(const zend_op *op, const char *name, int name_len, const char *cname, int cname_len, zval *r TSRMLS_DC);
void optimize_fullpath(zend_op *op, char *base, zend_bool use_include_path TSRMLS_DC);
int get_next_znode_pos(zend_op* op, zend_bool var_only);
void optimizer_parse_level(TSRMLS_D);
zend_uint optimizer_fetch_free_tmp_var(zend_op_array* op_array);

zend_op_array *optimize_op_array(zend_op_array *op_array, optimizer_source_file* sf TSRMLS_DC);
zend_op_array* optimizer_compile_file(zend_file_handle* h, int type TSRMLS_DC);


#if !((PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION > 2) || PHP_MAJOR_VERSION == 6) 

#	ifndef ALLOCA_FLAG
#		define ALLOCA_FLAG(name) zend_bool name;
#	endif

#	undef do_alloca
#	undef free_alloca
#	ifdef do_alloca_with_limit
#		define do_alloca(size, use_heap) do_alloca_with_limit(size, use_heap)
#		define free_alloca(p, use_heap) free_alloca_with_limit(p, use_heap)
#	else
#		define do_alloca(size, use_heap) emalloc(size)
#		define free_alloca(p, use_heap) efree(p)
#	endif

#endif

#endif /* OPTIMIZE_H*/
