/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2008 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Derick Rethans <derick@php.net>                             |
  |          Graham Kelly <graham@php.net>                               |
  +----------------------------------------------------------------------+
*/

/* $Id: php_optimizer.c,v 1.19 2009/04/30 03:32:27 myang Exp $ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "php_globals.h"
#include "zend.h"
#include "zend_API.h"
#include "zend_extensions.h"
#include "SAPI.h"
#include "ext/standard/info.h"

#include "php_optimizer.h"
#include "optimize.h"
#include "optimizer_debug.h"
#include "optimize_fcr.h"

#include <stdlib.h>


zend_compile_t *optimizer_old_compile_file;

ZEND_DECLARE_MODULE_GLOBALS(optimizer)


static PHP_INI_MH(OnUpdateInlineServer)
{
	OPTIMIZER_G(optimize_out_funcs_l) = new_value_length;
	return OnUpdateString(entry, new_value, new_value_length, mh_arg1, mh_arg2, mh_arg3, stage TSRMLS_CC);
}

static PHP_INI_MH(OnUpdateOptimizeOutFuncs)
{
	OPTIMIZER_G(optimize_inline_server_l) = new_value_length;
	return OnUpdateString(entry, new_value, new_value_length, mh_arg1, mh_arg2, mh_arg3, stage TSRMLS_CC);
}

PHP_INI_BEGIN()
	STD_PHP_INI_BOOLEAN("optimizer.enabled", "1", PHP_INI_SYSTEM, OnUpdateBool, optimizer_enabled, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.enable_cli", "1", PHP_INI_SYSTEM, OnUpdateBool, optimizer_enable_cli, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.enable_on_cache_only", "1", PHP_INI_SYSTEM, OnUpdateBool, optimizer_enable_cache, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_ENTRY("optimizer.level", "1", PHP_INI_SYSTEM, OnUpdateLong, optimization_level, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.test", "0", PHP_INI_SYSTEM, OnUpdateBool, optimizer_test, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.test_with_cache", "0", PHP_INI_SYSTEM, OnUpdateBool, optimizer_test_cache, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.verbose_debug", "0", PHP_INI_SYSTEM, OnUpdateBool, optimizer_verbose_debug, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.detect_static_function_calls", "0", PHP_INI_SYSTEM, OnUpdateBool, optimize_detect_static_function_calls, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.report_constants", "0", PHP_INI_SYSTEM, OnUpdateBool, optimize_report_constants, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.report_only_native_func", "0", PHP_INI_SYSTEM, OnUpdateBool, optimize_report_only_native_func, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.optimize_define", "1", PHP_INI_SYSTEM, OnUpdateBool, optimize_define, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.optimize_assert", "1", PHP_INI_SYSTEM, OnUpdateBool, optimize_assert, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_ENTRY("optimizer.inline_server", "", PHP_INI_SYSTEM, OnUpdateInlineServer, optimize_inline_server, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_BOOLEAN("optimizer.optimize_file_ops", "1", PHP_INI_SYSTEM, OnUpdateBool, optimize_file_ops, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_ENTRY("optimizer.optimize_out_funcs", "", PHP_INI_SYSTEM, OnUpdateOptimizeOutFuncs, optimize_out_funcs, zend_optimizer_globals, optimizer_globals)
	STD_PHP_INI_ENTRY("optimizer.disabled_functions", NULL, PHP_INI_SYSTEM, OnUpdateString, disabled_functions, zend_optimizer_globals, optimizer_globals)

PHP_INI_END()


static zend_bool optimizer_ini_bool_decode(zend_uint value_length, char* value)
{
	if ((value_length == 2 && strcasecmp("on", value) == 0) || 
		(value_length == 3 && strcasecmp("yes", value) == 0) || 
		(value_length == 4 && strcasecmp("true", value) == 0)) {
		return 1;
	}

	return 0;
}


/* {{{ optimizer_functions[]
 */
function_entry optimizer_functions[] = {
	{NULL, NULL, NULL}
};
/* }}} */



/* {{{ optimizer dependencies */
#if ZEND_MODULE_API_NO >= 20050922
zend_module_dep optimizer_module_deps[] = {
	ZEND_MOD_REQUIRED("standard")
#if __OPTIMIZER_USE_APC__
	ZEND_MOD_REQUIRED("apc")
#endif
	{NULL, NULL, NULL}
};
#endif
/* }}} */


/* {{{ optimizer_module_entry
 */
zend_module_entry optimizer_module_entry = {
#if ZEND_MODULE_API_NO >= 20050922
	STANDARD_MODULE_HEADER_EX, NULL,
	optimizer_module_deps,
#else
	STANDARD_MODULE_HEADER,
#endif
	"optimizer",
	optimizer_functions,
	PHP_MINIT(optimizer),
	PHP_MSHUTDOWN(optimizer),
	NULL,
	NULL,
	PHP_MINFO(optimizer),
	PHP_OPTIMIZER_VERSION,
	STANDARD_MODULE_PROPERTIES
};
/* }}} */


#ifdef COMPILE_DL_OPTIMIZER
ZEND_GET_MODULE(optimizer)
#endif


static void optimizer_globals_ctor(zend_optimizer_globals* optimizer_globals TSRMLS_DC) 
{
	//optimizer_globals->optimizer_enabled = 0;
	//optimizer_globals->optimizer_verbose_debug = 0;
	//optimizer_globals->optimize_assert = 1;
	//optimizer_globals->optimize_define = 0;
	//optimizer_globals->optimize_detect_static_function_calls = 0;
	//optimizer_globals->optimize_file_ops = 0;
	//optimizer_globals->optimize_inline_server = 0;
	//optimizer_globals->optimize_report_only_native_func = 0;
	//optimizer_globals->optimize_report_constants = 0;	
}

static void optimizer_globals_dtor(zend_optimizer_globals* optimizer_globals TSRMLS_DC) 
{
	/* Really, nothing to do here... but there probably will be (hash table???) */
}

/* {{{ PHP_MINIT_FUNCTION(optimizer) */
PHP_MINIT_FUNCTION(optimizer)
{
	ZEND_INIT_MODULE_GLOBALS(optimizer, optimizer_globals_ctor, optimizer_globals_dtor);
	REGISTER_INI_ENTRIES();
	
	OPTIMIZER_G(optimizer_flags) = 0;

	if (!OPTIMIZER_G(optimizer_enable_cli) && !strcmp(sapi_module.name, "cli")) {
		OPTIMIZER_G(optimizer_enabled) = 0;
	}

	zend_uint compile_f_reg = (OPTIMIZER_G(optimizer_enabled) ? OPTIMIZER_COMPILE_ZEND_REG : 0);
	zend_constant*   apc_magic     = NULL;

#if OPTIMIZER_DEBUG_COMPILE < 1
	OPTIMIZER_G(optimizer_verbose_debug) = 0;	
#endif
	
#if __OPTIMIZER_USE_APC__
	if (OPTIMIZER_G(optimizer_enabled)) {
		zend_bool try_apc = 1;
#  if OPTIMIZER_TEST_COMPILE
		if (OPTIMIZER_G(optimizer_test) && !OPTIMIZER_G(optimizer_test_cache)) {
			try_apc = 0;
		}
#  endif

		if (try_apc && zend_hash_find(EG(zend_constants), "\000apc_magic", sizeof("\000apc_magic"), (void**)&apc_magic) == SUCCESS) {
			if (OPTIMIZER_G(optimizer_enable_cache)) {
				zend_ini_entry* apc_enabled = NULL;

				if (zend_hash_find(EG(ini_directives), "apc.enabled", sizeof("apc.enabled"), (void**)&apc_enabled) == SUCCESS) {
					if (optimizer_ini_bool_decode(apc_enabled.value_length, apc_enabled.value)) {
						compile_f_reg = OPTIMIZER_COMPILE_APC_REG;
					}
				} else {
					compile_f_reg = OPTIMIZER_COMPILE_APC_REG;
				}
			}
		}
	}
#endif

	optimizer_old_compile_file = NULL;

	if (compile_f_reg == OPTIMIZER_COMPILE_ZEND_REG) {
		/* Register optimizer_compile_file() in ZEND */
		optimizer_old_compile_file = zend_compile_file;
		zend_compile_file = (zend_compile_t*)optimizer_compile_file;
		OPTIMIZER_G(optimizer_flags) |= OPTIMIZER_COMPILE_ZEND_REG;
	} else if (compile_f_reg == OPTIMIZER_COMPILE_APC_REG && apc_magic != NULL) {
		/* Register optimizer_compile_file() in APC */
		apc_set_compile_hook_t* setter = (apc_set_compile_hook_t*)Z_LVAL(apc_magic->value);
		optimizer_old_compile_file = setter((zend_compile_t*)optimizer_compile_file);
		OPTIMIZER_G(optimizer_flags) |= OPTIMIZER_COMPILE_APC_REG;
	}
	
	return SUCCESS;
}
/* }}} */



/* {{{ PHP_MSHUTDOWN_FUNCTION(optimizer) */
PHP_MSHUTDOWN_FUNCTION(optimizer)
{

	if (optimizer_old_compile_file != NULL) {
		if (OPTIMIZER_G(optimizer_flags) & OPTIMIZER_COMPILE_ZEND_REG) {
			zend_compile_file = optimizer_old_compile_file;
		}
	}

#ifndef ZTS
	optimizer_globals_dtor(&optimizer_globals TSRMLS_CC);
#else
    ts_free_id(optimizer_globals_id);
#endif	
  
  UNREGISTER_INI_ENTRIES();
  return SUCCESS;
}
/* }}} */



/* {{{ PHP_MINFO_FUNCTION(optimizer) */
PHP_MINFO_FUNCTION(optimizer)
{
	char buf[50];
	optimizer_parse_level(TSRMLS_C);
	
	php_info_print_table_start();
	if (OPTIMIZER_G(optimizer_enabled)) {
		php_info_print_table_row(2, "Opcode Optimizer", "enabled");
	} else {
		php_info_print_table_row(2, "Opcode Optimizer", "disabled");
	}
	
	
#if OPTIMIZER_DEBUG_COMPILE
	php_info_print_table_row(2, "Compiled with debug", "yes");
#else
	php_info_print_table_row(2, "Compiled with debug", "no");
#endif

#if OPTIMIZER_STATS_COMPILE
	php_info_print_table_row(2, "Compiled with stats", "yes");
#else
	php_info_print_table_row(2, "Compiled with stats", "no");
#endif

#if OPTIMIZER_TEST_COMPILE
	php_info_print_table_row(2, "Compiled with test suite", "yes");
#endif	
	
	
	if (OPTIMIZER_G(optimizer_verbose_debug)) {
		php_info_print_table_row(2, "Verbose debug", "on");
	} else {
		php_info_print_table_row(2, "Verbose debug", "off");
	}
	
	if (OPTIMIZER_G(optimizer_verbose_debug) && OPTIMIZER_G(optimizer_enabled)) {
		php_info_print_table_row(2, "compile_file hook", 
			(OPTIMIZER_G(optimizer_flags) & OPTIMIZER_COMPILE_ZEND_REG ? "Zend" : 
			(OPTIMIZER_G(optimizer_flags) & OPTIMIZER_COMPILE_APC_REG ? "APC" : "Unknown")));
	}

	OPTIMIZER_EXPERIMENTAL_BEGIN()
	if (OPTIMIZER_G(optimizer_verbose_debug)) {
		php_info_print_table_row(2, "Experimental Optimizations", "yes");
	}	
	OPTIMIZER_EXPERIMENTAL_END()

	php_info_print_table_row(2, "Version", PHP_OPTIMIZER_VERSION);
	
	sprintf(buf, "%d", OPTIMIZER_G(optimization_level));
	php_info_print_table_row(2, "Optimization Level", buf);
	
	sprintf(buf, "%d", OPTIMIZER_G(peephole_passes));
	php_info_print_table_row(2, "Peephole Passes", buf);
	
	php_info_print_table_end();
	DISPLAY_INI_ENTRIES();
}
/* }}} */



/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
