/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2008 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Derick Rethans <derick@php.net>                             |
  |          Graham Kelly <graham@php.net>                               |
  +----------------------------------------------------------------------+
*/

/* $Id: php_optimizer.h,v 1.16 2009/04/30 03:32:27 myang Exp $ */

#ifndef PHP_OPTIMIZER_H
#define PHP_OPTIMIZER_H

#include "php.h"
#include "zend.h"
#include "zend_API.h"
#include "zend_extensions.h"

#define PHP_OPTIMIZER_VERSION "0.0.1"

/*
 * If set to 1 the debug information and functions will be compiled into
 *   the optimizer. This is a compile time switch only. Therefore debug mode
 *   may still be off if optimizer.verbose_debug does not eual 1 in the php.ini
 * 
 */
#ifndef OPTIMIZER_DEBUG_COMPILE
#  define OPTIMIZER_DEBUG_COMPILE 1
#endif
              

/*
 * If set to 1 statistics on how well the optimizer performed in optimizing 
 *   will be tabulated and available for output to the user.
 * 
 */
#ifndef OPTIMIZER_STATS_COMPILE  
#  define OPTIMIZER_STATS_COMPILE 1
#endif


/*
 * Id set to 1 test code will be compiled in. 
 *
 */
#ifndef OPTIMIZER_TEST_COMPILE
#  define OPTIMIZER_TEST_COMPILE 1
#endif


#ifndef __OPTIMIZER_USE_APC__
#  define __OPTIMIZER_USE_APC__ 0
#endif

#ifndef __OPTIMIZER_EXPERIMENTAL__
#  define __OPTIMIZER_EXPERIMENTAL__ 1
#endif

extern zend_module_entry optimizer_module_entry;
#define phpext_optimizer_ptr &optimizer_module_entry


/*
 * For compatability reasons ONLY... this is just a simple fix to the implicit function delaration problem
 *  
 */
typedef zend_op_array* (*apc_optimize_function_t) (zend_op_array* TSRMLS_DC);
extern apc_optimize_function_t apc_register_optimizer(apc_optimize_function_t optimizer TSRMLS_DC);


#define OPTIMIZER_COMPILE_ZEND_REG  1 /* optimizer_compile_file() registered with ZendEngine */
#define OPTIMIZER_COMPILE_APC_REG   2 /* optimizer_compile_file() registered with APC */

/*
 * Optimizer Globals
 *   Naming conventions: 
 *     The optimizer_* globals tell the optimizer itself how to function
 *     The optimize_* globals indicate how to (or if to) perform a specific optimization (type)
 */
ZEND_BEGIN_MODULE_GLOBALS(optimizer)
	zend_bool optimizer_enabled; /* if true then the optimizer is enabled and we run it! */
	zend_bool optimizer_enable_cli;
	zend_bool optimizer_enable_cache;
	zend_bool optimizer_test; /* if true the optimizer will print internal test data */
	zend_bool optimizer_test_cache;
	zend_bool optimizer_verbose_debug; /* if true the optimizer will print a LOT of debug information */
	zend_bool optimize_detect_static_function_calls; /* if true then the optimizer will report when it finds any function calls with all static params */
	zend_bool optimize_report_constants; /* if true then optimizer will ??? */
	zend_bool optimize_report_only_native_func; /* if true the optimizer will ??? */
	zend_bool optimize_define; /* if true the optimizer will ??? (maybe optimize define's?) */
	zend_bool optimize_assert; /* if true the optimizer will attempt to optimize asserts */
	char*     optimize_inline_server;   /* if not null the optimizer will attempt to remove any _SERVER vars with keys matching the names in the string */
	zend_uint optimize_inline_server_l;
	zend_bool optimize_file_ops; /* if true the optimizer will optimize file opperations */
  
	char* optimize_out_funcs;
	zend_uint optimize_out_funcs_l;
  
  zend_uint optimization_level; /* An arbitrary level that specifies what will happen in the optimizer */
  zend_uint peephole_passes;    /* The number of passes to be done by the peephole optimizer */
  zend_bool optimize_for_speed; /* true if the optimizer should favor execution time speed over memory conservation */
  char*     disabled_functions; /*** space separated list of functions for which optimization is disabled ***/
  zend_op_array *op_array;      /*** op_array being processed ***/
                                /* We will try to optimize for both as much as possible but when a decission for one over the other needs to be made this will help decide it */

	zend_uint optimizer_flags;
ZEND_END_MODULE_GLOBALS(optimizer)

ZEND_EXTERN_MODULE_GLOBALS(optimizer)


#ifdef ZTS
#  include "TSRM.h"
#  define OPTIMIZER_G(v) TSRMG(optimizer_globals_id, zend_optimizer_globals*, v)
#else
#  define OPTIMIZER_G(v) (optimizer_globals.v)
#endif

#define OPTIMIZER_APC_USED() (OPTIMIZER_G(optimizer_flags) & OPTIMIZER_COMPILE_APC_REG)


#if __OPTIMIZER_EXPERIMENTAL__
#  define OPTIMIZER_EXPERIMENTAL_BEGIN() if (OPTIMIZER_G(optimization_level) >= 6) {
#  define OPTIMIZER_EXPERIMENTAL_END()   }
#else
#  define OPTIMIZER_EXPERIMENTAL_BEGIN() if (0) {
#  define OPTIMIZER_EXPERIMENTAL_END()   }
#endif

/*
 * Startup, Shutdown, Info functions
 * 
 */
PHP_MINIT_FUNCTION(optimizer);
PHP_MSHUTDOWN_FUNCTION(optimizer);
PHP_MINFO_FUNCTION(optimizer);

typedef zend_op_array* (zend_compile_t)(zend_file_handle*, int TSRMLS_DC);
typedef zend_compile_t* (apc_set_compile_hook_t)(zend_compile_t*);
extern zend_compile_t *optimizer_old_compile_file;

#endif	/* PHP_OPTIMIZER_H */

/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * indent-tabs-mode: t
 * End:
 */
